import { Provide } from '@midwayjs/decorator';
import { InjectEntityModel } from '@midwayjs/typegoose';
import { ReturnModelType } from '@typegoose/typegoose';
import {
  BaseService,
  FrameworkErrorEnum,
  ServiceError,
} from 'free-midway-component';
import { CpcMessage } from '../entity/cpcMessage.entity';
import { Config } from '@midwayjs/core';
import _ = require('lodash');
const compressing = require('compressing');
const cOnvert = require('xml-js');
import * as fs from 'node:fs';
import * as Path from 'node:path';
const moment = require('moment');
import { Apply } from '../entity/apply.entity';
import { Patent } from '../entity/patent.entity';
import { CpcMessageerror } from '../entity/cpcMessageerror.entity';
const sep = Path.sep;
import { Types } from 'mongoose';
const ObjectId = Types.ObjectId;
type modelType = ReturnModelType<typeof CpcMessage>;
@Provide()
export class CpcMessageService extends BaseService<modelType> {
  @InjectEntityModel(CpcMessage)
  model: modelType;

  @InjectEntityModel(Apply)
  ApplyModel: ReturnModelType<typeof Apply>;

  @InjectEntityModel(Patent)
  PatentModel: ReturnModelType<typeof Patent>;

  @InjectEntityModel(CpcMessageerror)
  errorModel: ReturnModelType<typeof CpcMessageerror>;

  @Config('export.root_path')
  import_root_path;

  async import(body) {
    let uri = body.url;
    uri = uri.replace('/files', this.import_root_path);
    const fileFullName = Path.basename(uri);
    const arr = fileFullName.split('.');
    const ext = _.last(arr);
    const uncompressFilePath = `${this.import_root_path}${sep}temp`;
    if (ext !== 'zip') {
      throw new ServiceError(
        '只接收压缩格式为zip的压缩包',
        FrameworkErrorEnum.SERVICE_FAULT
      );
    }
    // 解压
    try {
      await compressing.zip.uncompress(uri, `${uncompressFilePath}`);
    } catch (error) {
      throw new ServiceError('解压失败', FrameworkErrorEnum.REQUEST_FAULT);
    }
    // 修改,多个专利:需要扫描下解压后的文件夹,然后循环处理下面所有的内容
    const tempDirRootPath = `${uncompressFilePath}${sep}`;
    let tempDir = fs.readdirSync(tempDirRootPath);
    const errorList = [];
    // 2023-02-06: zip格式修改: zip套dir => zip套zip, zip解压出dir;
    // 也就是需要将zip再次解压一遍,并将所有的zip删除即可
    const isAllZip = tempDir.some(f => f.includes('zip'));
    if (isAllZip) {
      for (const p of tempDir) {
        if (!p.includes('.zip')) continue;
        const zipFullPath = `${uncompressFilePath}${sep}${p}`;
        const zipArr = p.split('.');
        const name = _.head(zipArr);
        try {
          await compressing.zip.uncompress(
            zipFullPath,
            `${uncompressFilePath}${sep}${name}`
          );
        } catch (error) {
          await this.dirDelete(uncompressFilePath);
          errorList.push({
            patent_key: p,
            content: '內部解压包,解压发生错误',
          });
        }
      }
      // 删除zip
      await this.deleteZip(uncompressFilePath);
      tempDir = fs.readdirSync(tempDirRootPath);
    }
    for (const tempDirPath of tempDir) {
      if (tempDirPath.includes('.zip')) continue;
      const thisDirPath = Path.resolve(tempDirRootPath, tempDirPath);
      const patentNumber = tempDirPath;
      // 找到
      let filePaths = [];
      try {
        filePaths = await this.findFileNameLoop(thisDirPath);
      } catch (error) {
        errorList.push({
          patent_key: patentNumber,
          content: '格式不对',
        });
        continue;
      }
      if (filePaths.length <= 0) {
        errorList.push({ patent_key: patentNumber, content: '未找到文件' });
        continue;
      }
      // 2021-10-16 修改: 取出第一层的xml文件,即:路径最短的;除文件名外,字符串长度最小的
      let xmlPaths = filePaths.filter(f => f.includes('.xml'));
      if (xmlPaths.length > 0) {
        // 去掉文件名
        xmlPaths = xmlPaths.map(i => {
          const arr = i.split(sep);
          arr.pop();
          const obj = { dir: arr.join(sep), fullPath: i };
          return obj;
        });
        xmlPaths = xmlPaths.filter(f => f.fullPath.includes('list.xml'));
      }
      const xmlPath = _.get(_.head(xmlPaths), 'fullPath');
      if (!xmlPath) {
        errorList.push({ patent_key: patentNumber, content: '未找到xml文件' });
        continue;
      }
      let xmlData;
      try {
        xmlData = await this.dealXml(xmlPath);
      } catch (error) {
        errorList.push({
          patent_key: patentNumber,
          content: `解析xml数据中的 申请号 和 专利名称 失败 于:${xmlPath}`,
          ...xmlData,
        });
        continue;
      }
      if (!xmlData) {
        errorList.push({
          patent_key: patentNumber,
          content: '未找到xml的数据',
        });
        continue;
      }
      /**
       * @property {Array} patent 专利信息
       * @property {Boolean || null} needChange 是否修改数据
       */
      let result;
      try {
        result = await this.getPatent(xmlData);
      } catch (error) {
        errorList.push({
          patent_key: patentNumber,
          content: '查询专利信息失败',
          ...xmlData,
        });
        continue;
      }
      // 2022-09-02 数组处理模式 ↑
      const patents = _.get(result, 'patent');
      if (!patents || patents.length <= 0) {
        errorList.push({
          patent_key: patentNumber,
          content: '未通过 申请号 或 专利名称 找到指定专利',
          ...xmlData,
        });
        continue;
      }
      for (const patent of patents) {
        // 2022-09-02 原处理进入循环中处理
        let warningData;
        try {
          warningData = await this.getWarningData(patent);
        } catch (error) {
          errorList.push({
            patent_key: patentNumber,
            content: '组织专利信息警告数据失败',
            ...xmlData,
          });
          continue;
        }
        // 2023-02-07 原来是.tif,现在是.pdf
        const uploadPaths = filePaths.filter(f => f.includes('.pdf'));
        if (uploadPaths && uploadPaths.length > 0) {
          try {
            await this.toUploadFile(warningData, uploadPaths);
          } catch (error) {
            errorList.push({
              patent_key: patentNumber,
              content: '存储数据失败',
              ...xmlData,
            });
            continue;
          }
        }
      }
    }
    try {
      await this.dirDelete(uncompressFilePath);
    } catch (error) {
      this.ctx.logger.error('清除缓存失败');
    }
    // 将错误信息添加到表中
    const errorMsg = errorList.map(i => ({
      create_number: i.create_number || '',
      patent_name: i.name || '',
      patent_key: i.patent_key,
      content: i.content,
      zip_url: uri,
      create_date: moment().format('YYYY-MM-DD HH:mm:ss'),
    }));
    await this.errorModel.insertMany(errorMsg);
    if (errorList.length > 0) return errorList;
    return 'ok';
  }

  /**
   * 上传图片,保存
   * @param {Array} data 警告数据,未保存
   * @param {Array} files tif图片文件路径数组
   */
  async toUploadFile(data, files) {
    const filePaths = [];
    let name = '';
    const savePath = new ObjectId();
    for (const file of files) {
      let arr = file.split(sep);
      const fullFileName: any = _.last(arr);
      arr = fullFileName.split('.');
      // 文件后缀
      const ext = _.last(arr);
      // 图片文件
      const buffer = await fs.readFileSync(file);
      // 新文件名
      const randomStr = new ObjectId();
      // 新文件名带后缀
      const newFileName = `${randomStr}.${ext}`;
      name = newFileName;
      // 上传后部分路径
      const shortPath = `platform${sep}patent${sep}${savePath}${sep}`;
      // 上传全路径
      const path = `${this.import_root_path}${sep}${shortPath}`;
      // 检查上传全路径是否存在,不存在就创建文件夹
      await this.checkPathAndCreate(path);
      // 另存文件
      await fs.writeFileSync(`${path}${newFileName}`, buffer);
      // 替换为files开头
      let dataPath = path.replace(this.import_root_path, '/files');
      while (dataPath.indexOf('\\') >= 0) {
        dataPath = dataPath.replace('\\', '/');
      }
      // 补全请求用路径的所有路径
      dataPath += newFileName;
      filePaths.push(dataPath);
    }
    for (const i of data) {
      i.warn_file = {
        name: name,
        uri: filePaths,
      };
      i.send_date = moment().format('YYYY-MM-DD');
    }
    const res = await this.model.insertMany(data);
    return res;
  }

  /**
   * 检查路径,没有创建
   * @param {Array} filePath 图片路径数组
   */
  async checkPathAndCreate(filePath) {
    const arr = filePath.split(sep);
    // w:第一个是盘符,直接和第二个合并
    // l:前几个应该都是存在的.也可以合并
    let head = _.head(arr);
    if (arr[1]) head = `${head}${sep}${arr[1]}`;
    arr.shift();
    arr[0] = head;
    let curPath;
    for (const p of arr) {
      curPath = `${curPath ? curPath + sep : ''}${p}`;
      if (!fs.existsSync(curPath)) {
        await fs.mkdirSync(curPath);
      }
    }
  }

  /**
   * 组织专利预警数据
   * @param {Object} patent 专利
   */
  async getWarningData(patent) {
    const object: any = {};
    const arr = [];
    if (_.get(patent, 'create_number'))
      object.create_number = _.get(patent, 'create_number');
    if (_.get(patent, '_id')) object.patent_id = _.get(patent, '_id');
    if (_.get(patent, 'name')) object.patent_name = _.get(patent, 'name');
    // 找人的信息
    const users = _.get(patent, 'users');
    if (users && _.isArray(users)) {
      for (const user of users) {
        if (!_.get(user, 'user_id') || !_.get(user, 'name')) continue;
        const { user_id, name } = user;
        const obj = _.cloneDeep(object);
        obj.receive_id = user_id;
        obj.receive_name = name;
        arr.push(obj);
      }
      // 2021-11-18 添加,user,mech,admin,agentmech是否存在,如果存在,也添加预警信息
      const list = ['user', 'mech', 'admin', 'agentmech'];
      const suffix = ['id', 'name'];
      for (const prefix of list) {
        const marr = suffix.map(i => `${prefix}_${i}`);
        const r = marr.every(i => _.get(patent, i));
        if (r) {
          const obj = _.cloneDeep(object);
          for (const key of suffix) {
            obj[`to_${key}`] = _.get(patent, `${prefix}_${key}`);
          }
          arr.push(obj);
        }
      }
    } else if (users) {
      const obj = _.cloneDeep(object);
      const name = _.get(patent, 'user_name');
      obj.receive_id = users;
      obj.receive_name = name;
      arr.push(obj);
    }
    return arr;
  }

  /**
   * 在patentApply/patentInfo表中找指定的专利
   * @param {Object} object xml的数据
   */
  async getPatent(object) {
    let { create_number } = object;
    const name = object.name;
    create_number = `CN${create_number}`;
    const arr = create_number.split('');
    arr.splice(arr.length - 1, 0, '.');
    create_number = arr.join('');
    const needChange = false;
    // 改为多条数据查询,不再针对一个数据
    // 1,名称相同,关联人不同:
    const listPA = await this.ApplyModel.find({
      $or: [{ create_number }, { name }],
    });
    for (const i of listPA) {
      const { create_number: ocn, _id } = i;
      // create_number 不一致,就去更改
      if (ocn !== create_number)
        await this.ApplyModel.updateOne({ _id }, { create_number });
    }
    const listPI = await this.PatentModel.find({
      $or: [{ create_number }, { name }],
    });
    const res = [...listPA, ...listPI];
    const returnObject: any = {};
    if (res.length > 0) {
      returnObject.patent = JSON.parse(JSON.stringify(res));
      if (needChange) {
        returnObject.create_number = create_number;
        returnObject.needChange = needChange;
      }
      return returnObject;
    }
    return null;
  }

  /**
   * 根据路径找到xml,再找到该xml的申请号和专利名称
   * @param {String} filePath xml的文件路径
   */
  async dealXml(filePath) {
    const buffer = await fs.readFileSync(filePath, { encoding: 'utf8' });
    if (!buffer)
      throw new ServiceError(
        '读取xml文件失败',
        FrameworkErrorEnum.REQUEST_FAULT
      );
    const json = JSON.parse(
      cOnvert.xml2json(buffer, { compact: true, spaces: 4 })
    );
    const create_number = _.get(
      json,
      'data-bus.TONGZHISXJ.SHUXINGXX.SHENQINGH._text'
    );
    const name = _.get(json, 'data-bus.TONGZHISXJ.SHUXINGXX.FAMINGMC._text');
    if (!name) return null;
    return { name, create_number };
  }

  /**
   * 找到深层的tif和xml
   * 因为命名方式都是以文件名,之后tif的文件夹也在深层fileName.套娃且文件名都是fileName
   * @param {String} uri temp路径
   */
  async findFileNameLoop(uri) {
    const filePath = uri;
    if (!fs.existsSync(filePath)) {
      return [];
    }
    const files = fs.readdirSync(filePath);
    let result = [];
    for (const file of files) {
      const curPath = Path.resolve(filePath, file);
      if (fs.statSync(curPath).isDirectory()) {
        const arr = await this.findFileNameLoop(curPath);
        result.push(...arr);
      } else {
        result.push(curPath);
      }
    }
    //现在是.pdf
    result = result.filter(f => f.includes('.pdf') || f.includes('.xml'));
    return result;
  }

  /**
   * 删除路径下的所有内容
   * @param {String} filePath 路径
   */
  async dirDelete(filePath) {
    if (fs.existsSync(filePath)) {
      const files = fs.readdirSync(filePath);
      for (const file of files) {
        const curPath = `${filePath}${sep}${file}`;
        if (fs.statSync(curPath).isDirectory()) {
          this.dirDelete(curPath);
        } else {
          fs.unlinkSync(curPath);
        }
      }
      fs.rmdirSync(filePath);
    }
  }
  /**
   * 删除路径下的所有zip
   * @param {String} filePath 路径
   */
  async deleteZip(filePath) {
    if (!fs.existsSync(filePath)) return;
    const files = fs.readdirSync(filePath);
    for (const file of files) {
      if (!file.includes('.zip')) continue;
      const curPath = `${filePath}${sep}${file}`;
      if (!fs.existsSync(curPath)) continue;
      fs.unlinkSync(curPath);
    }
  }
}
